import ts from 'typescript';
import path from 'path';
import fs from 'fs';
import nodeResolve from '@rollup/plugin-node-resolve';

const PLUGIN_NAME = "ts-paths";
const tsPathsResolve = ({
  tsConfigPath = process.env["TS_NODE_PROJECT"] || ts.findConfigFile(".", ts.sys.fileExists) || "tsconfig.json",
  logLevel = "warn",
  fallback = nodeResolve({
    extensions: [".mjs", ".js", ".json", ".node", ".jsx", ".ts", ".tsx"]
  })
} = {}) => {
  if (logLevel === "debug") {
    console.log(formatLog("info", PLUGIN_NAME, `typescript version: ${ts.version}`));
  }

  let compilerOptions = getTsConfig(tsConfigPath, PLUGIN_NAME);
  let mappings = createMappings({
    paths: compilerOptions.paths,
    logLevel,
    pluginName: PLUGIN_NAME
  });
  return {
    name: PLUGIN_NAME,

    buildStart() {
      compilerOptions = getTsConfig(tsConfigPath, PLUGIN_NAME);
      mappings = createMappings({
        paths: compilerOptions.paths,
        logLevel,
        pluginName: PLUGIN_NAME
      });
      return null;
    },

    async resolveId(request, importer) {
      if (typeof importer === "undefined" || request.startsWith("\0")) {
        return null;
      }

      const [resolved, nodeModules] = findResolve({
        compilerOptions,
        mappings,
        request,
        importer
      });

      if (resolved) {
        if (logLevel === "debug") {
          console.log(formatLog("info", PLUGIN_NAME, `${request} -> ${resolved}`));
        }

        if (nodeModules) {
          return fallback.resolveId.bind(this)(resolved, importer, {});
        }

        return resolved;
      }

      return fallback.resolveId.bind(this)(request, importer, {});
    }

  };
};

function formatLog(level, pluginName, value) {
  switch (level) {
    case "error":
      return `\x1b[1;31m[${pluginName}]: ${value}\x1b[0m`;

    case "warn":
      return `\x1b[1;33m[${pluginName}]: ${value}\x1b[0m`;

    default:
      return `\x1b[1;34m[${pluginName}]: ${value}\x1b[0m`;
  }
}

const getTsConfig = (tsConfigPath, pluginName) => {
  const {
    error,
    config
  } = ts.readConfigFile(tsConfigPath, ts.sys.readFile);

  if (error) {
    throw new Error(formatLog("error", pluginName, error.messageText));
  }

  let {
    options: compilerOptions
  } = ts.parseJsonConfigFileContent(config, ts.sys, path.dirname(tsConfigPath));

  if (!compilerOptions) {
    throw new Error(formatLog("error", pluginName, "'compilerOptions' is gone."));
  }

  if (compilerOptions.baseUrl == undefined) {
    compilerOptions.baseUrl = path.dirname(tsConfigPath);
  }

  if (!compilerOptions.paths || compilerOptions.paths instanceof Array) {
    compilerOptions.paths = {};
  }

  return compilerOptions;
};

function createMappings({
  paths,
  logLevel = "none",
  pluginName = PLUGIN_NAME
}) {
  const countWildcard = value => value.match(/\*/g)?.length;

  const valid = value => /(\*|\/\*|\/\*\/)/.test(value);

  const mappings = [];

  for (const pattern of Object.keys(paths)) {
    if (countWildcard(pattern) > 1) {
      logLevel != "none" && console.warn(formatLog("warn", pluginName, `Pattern '${pattern}' can have at most one '*' character.`));
      continue;
    }

    const wildcard = pattern.indexOf("*");

    if (wildcard !== -1 && !valid(pattern)) {
      logLevel != "none" && console.warn(formatLog("warn", pluginName, `path pattern '${pattern}' is not valid.`));
      continue;
    }

    const targets = paths[pattern].filter(target => {
      if (countWildcard(target) > 1) {
        logLevel != "none" && console.warn(formatLog("warn", pluginName, `Substitution '${target}' in pattern '${pattern}' can have at most one '*' character.`));
        return false;
      }

      const wildcard = target.indexOf("*");

      if (wildcard !== -1 && !valid(target)) {
        logLevel != "none" && console.warn(formatLog("warn", pluginName, `target pattern '${target}' is not valid`));
        return false;
      }

      if (target.indexOf("@types") !== -1 || target.endsWith(".d.ts")) {
        logLevel != "none" && console.warn(formatLog("warn", pluginName, `type defined ${target} is ignored.`));
        return false;
      }

      return true;
    });

    if (targets.length == 0) {
      continue;
    }

    if (pattern === "*") {
      mappings.push({
        wildcard: true,
        pattern,
        prefix: "",
        suffix: "",
        targets
      });
      continue;
    }

    mappings.push({
      wildcard: wildcard !== -1,
      pattern,
      prefix: pattern.substr(0, wildcard),
      suffix: pattern.substr(wildcard + 1),
      targets
    });
  }

  if (logLevel === "debug") {
    for (const mapping of mappings) {
      console.log(formatLog("info", pluginName, `pattern: '${mapping.pattern}' targets: '${mapping.targets}'`));
    }
  }

  return mappings;
}

function isPatternMatch(prefix, suffix, candidate) {
  return candidate.length >= prefix.length + suffix.length && candidate.startsWith(prefix) && candidate.endsWith(suffix);
}

function findMatch(request, mappings) {
  let longestMatchedPrefixLength = 0;
  let matched;

  for (const mapping of mappings) {
    const {
      wildcard,
      prefix,
      suffix,
      pattern: source
    } = mapping;

    if (wildcard && isPatternMatch(prefix, suffix, request)) {
      if (longestMatchedPrefixLength < prefix.length) {
        longestMatchedPrefixLength = prefix.length;
        matched = mapping;
      }
    } else if (request === source) {
      matched = mapping;
      break;
    }
  }

  return matched;
}

const findResolve = ({
  mappings,
  request,
  importer,
  compilerOptions
}) => {
  const matched = findMatch(request, mappings);

  if (!matched) {
    return ["", false];
  }

  const matchedWildcard = request.slice(matched.prefix.length, request.length - matched.suffix.length);

  for (const target of matched.targets) {
    let predicted = target;

    if (matched.wildcard) {
      predicted = target.replace("*", matchedWildcard);
    }

    const answer = path.resolve(compilerOptions.baseUrl, predicted);

    if (answer.indexOf("node_modules/") != -1) {
      return [answer, true];
    } // NOTE: resolve module path with typescript API


    const result = ts.resolveModuleName(answer, importer, compilerOptions, ts.sys);

    if (result?.resolvedModule) {
      return [result.resolvedModule.resolvedFileName, false];
    } // NOTE: For those are not modules, ex: css, fonts...etc.


    if (fs.existsSync(answer)) {
      return [answer, false];
    }
  }

  return ["", false];
};

export { createMappings, tsPathsResolve as default, findMatch, tsPathsResolve };
//# sourceMappingURL=index.esm.js.map
